import json
import prolysis.rules_handling.declare_processing as declare_processing


def rules_from_json(file_path):
    with open(file_path, 'r') as file:
        declare_file = json.load(file)
    return declare_file['constraints'],set(declare_file['tasks'])
def preprocess(rules):
    rules_processes = []
    co_exist_list = []
    absence_list = set([r['parameters'][0][0] for r in rules if r['template'] == "Absence"])
    for r in rules:
        if r['template'] == 'AtLeast2' or r['template'] == 'AtLeast3':
            r_new = r.copy()
            r_new['template'] = 'AtLeast1'
            rules_processes.append(r_new)
        elif r['template'] =="Absence":
            continue
        elif r['template'] == 'AtMost2' or r['template'] == 'AtMost3':
            continue
        elif r['template'] == 'CoExistence':
            co_exist_list.append((r['parameters'][0][0],r['parameters'][1][0]))
            if (r['parameters'][1][0],r['parameters'][0][0]) not in co_exist_list:
                rules_processes.append(r)
        else:
            rules_processes.append(r)
    return rules_processes,absence_list

def dfa_list_generator(rules,S_mapped):
    dfa_list = []
    for c in rules:
        if "Absence" in c['template'] or "Init" in c['template'] or "End" in c['template'] or "AtMost" in c[
            'template'] or "AtLeast" in c['template']:
            dfa_to_add = declare_processing.gen_reg_dfa(c['template'], [c['parameters'][0][0]], S_mapped)
            dfa_list.append(((c['template'], (c['parameters'][0][0])), dfa_to_add, c['support'],c['confidence']))
        else:
            dfa_to_add = declare_processing.gen_reg_dfa(c['template'],
                                                        [c['parameters'][0][0], c['parameters'][1][0]],
                                                        S_mapped)
            dfa_list.append((
                            (c['template'], (c['parameters'][0][0], c['parameters'][1][0])),
                            dfa_to_add, c['support'],c['confidence']))

    total_sup = sum([el[2] for el in dfa_list])
    total_conf = sum([el[3] for el in dfa_list])
    return dfa_list, total_sup, total_conf




import pandas as pd
import matplotlib
matplotlib.use('agg')
import scipy.signal as sci_sig
import os
import itertools
import pm4py
import json
import ast
from prolysis.analysis.python_emsc import stochastic
import plotly.graph_objects as go
import numpy as np
from prolysis.util.redis_connection import redis_client
from prolysis.util.logging import log_command

# # Constants
OUTPUT_DIR = "output_files"

os.makedirs(OUTPUT_DIR, exist_ok=True)  # Ensure output directory exists
color_theme_drift_map = 'Blues'


def emd_dist(bin1, bin2):
    if isinstance(bin1, list):
        bin1 = pd.Series(bin1)
    if isinstance(bin2, list):
        bin2 = pd.Series(bin2)
    lang1 = bin1.value_counts(normalize=True)
    lang1 = {ast.literal_eval(key): value for key, value in lang1.items()}
    lang2 = bin2.value_counts(normalize=True)
    lang2 = {ast.literal_eval(key): value for key, value in lang2.items()}
    dist = round(stochastic.compare_languages_levenshtein(lang1, lang2),2)
    return dist

def bins_generation(kpi, n_bin):
    """Generate bins and map ranges for a given KPI."""
    case_table = pd.read_csv("output_files/out.csv").sort_values(by=[kpi])

    map_range = {i: float(case_table[kpi].iloc[round((i / n_bin) * len(case_table[kpi]))]) for i in range(n_bin)}
    map_range[n_bin] = float(case_table[kpi].iloc[-1])

    bin_size = round(len(case_table) / n_bin)
    bins = [
        ((min(case_table[point:point + bin_size][kpi]), max(case_table[point:point + bin_size][kpi])), idx,
         case_table[point:point + bin_size]['trace'])
        if point + bin_size < len(case_table) else
        ((min(case_table[point:][kpi]), max(case_table[point:][kpi])), idx, case_table[point:]['trace'])
        for idx, point in enumerate(range(0, len(case_table), bin_size))
    ]

    return bins, map_range, case_table


def sliding_window(bins, n_bin,WINDOWS):
    """Perform sliding window analysis for change detection."""
    df = pd.DataFrame(0.0, index=WINDOWS, columns=[i for i in range(1, n_bin)])
    for window_size in WINDOWS:
        for mid in range(window_size, n_bin - window_size + 1):
            left = [item for b in bins[mid - window_size:mid] for item in b[2]]
            # left = combine_dics([b[2] for b in bins[mid - window_size:mid]])
            right = [item for b in bins[mid:mid + window_size] for item in b[2]]
            # right = combine_dics([b[2] for b in bins[mid:mid + window_size]])
            df.at[window_size, mid] = emd_dist(left,right)
    masks = [
        [True] * (window - 1) + [False] * (n_bin - 2 * window + 1) + [True] * (window - 1)
        for window in WINDOWS
    ]
    return df, masks


def segmentation(df,bins,n_bin,w,sig):
    peaks, _ = sci_sig.find_peaks(df.loc[w], height=[sig])
    peaks = [int(p) for p in peaks]
    segments = []
    segments_ids = []
    last_p = -1
    x_state = 0
    for p in peaks:
        new = (x_state, x_state, pd.Series(list(itertools.chain.from_iterable([x[2] for x in bins[last_p + 1:p + 1]]))))
        # new_ids = [item for b in bins[last_p + 1:p + 1] for item in b[2].index]
        new_ids = [item for b in bins[last_p + 1:p + 1] for item in b[2].index]
        segments.append(new)
        segments_ids.append(new_ids)
        last_p = p
        x_state += 1

    new = (x_state, x_state, pd.Series(list(itertools.chain.from_iterable([x[2] for x in bins[last_p + 1:]]))))
    new_ids = [item for b in bins[last_p + 1:] for item in b[2].index]
    segments.append(new)
    segments_ids.append(new_ids)

    state = n_bin
    m_dend = []
    cal_list = {}

    itr = 1
    data_points = []

    matrices = []
    labels = []
    mins_vec = []

    state_dic = {}

    while len(segments) > 1:
        dist_matrix = np.ones((len(segments), len(segments)))
        for i in range(0, len(segments)):
            for j in range(i + 1, len(segments)):
                if (segments[i][0], segments[j][0]) not in cal_list:
                    cal_list[(segments[i][0], segments[j][0])] = emd_dist(segments[i][2], segments[j][2])
                dist_matrix[i, j] = cal_list[(segments[i][0], segments[j][0])]
                data_points.append((itr, dist_matrix[i, j]))
        matrices.append(dist_matrix)
        labels.append([b[1] for b in segments])
        min_dist_ind = np.unravel_index(np.argmin(dist_matrix, axis=None), dist_matrix.shape)
        mins_vec.append(dist_matrix[min_dist_ind])
        m_dend.append([segments[min_dist_ind[0]][1], segments[min_dist_ind[1]][1], dist_matrix[min_dist_ind],
                       len(segments[min_dist_ind[0]][2]) + len(segments[min_dist_ind[1]][2])])
        state_dic[state] = ((segments[min_dist_ind[0]][1], len(segments[min_dist_ind[0]][2])),
                            (segments[min_dist_ind[1]][1], len(segments[min_dist_ind[1]][2])))
        segments = [segments[k] for k in range(0, len(segments)) if (k != min_dist_ind[0] and k != min_dist_ind[1])] + [
            ((segments[min_dist_ind[0]][0], segments[min_dist_ind[1]][0]), state,
             pd.concat([segments[min_dist_ind[0]][2], segments[min_dist_ind[1]][2]]))]
        state = state + 1
        itr += 1


    ittr = 0

    new_m = matrices[ittr]
    for i in range(0, len(matrices[ittr])):
        for j in range(i, len(matrices[ittr][i])):
            if i == j:
                new_m[j, i] = 0
            else:
                new_m[j, i] = new_m[i, j]
    return segments, segments_ids, new_m, peaks


def plot_figures_EMD(df, masks, n_bin, map_range,WINDOWS):

    every = 1

    # Replace masked values with NaN for transparency
    heatmap_data = np.ma.array(df, mask=masks).filled(np.nan)

    # Generate x-tick labels
    x_labels = [
        f"{round(x * (100 / n_bin))}% ({round(map_range[x], 1)})"
        for x in range(1, n_bin, every)
    ]

    # Generate y-tick labels
    y_labels = WINDOWS

    # Define the colorscale (exclude gray, focus on valid data range)
    custom_colorscale = "Reds"  # Default Plotly Reds colorscale

    # Create the Plotly Heatmap
    fig1 = go.Figure(
        data=go.Heatmap(
            x=x_labels,
            z=heatmap_data,  # Heatmap values
            y=y_labels,
            colorscale=custom_colorscale,  # Valid data colorscale
            zmin=np.nanmin(df),  # Min value for color scaling
            zmax=np.nanmax(df),  # Max value for color scaling
            colorbar=dict(title="ldist"),  # Colorbar title
            showscale=True,  # Show the scale
        )
    )


    fig1.update_layout(
        # paper_bgcolor="gray",  # Set the figure's background color to gray
        plot_bgcolor="gray",  # Set the plot area background color to gray
        title=dict(
            text="Sliding Window Analysis",  # Title text
            x=0.5,  # Horizontal alignment (0=left, 0.5=center, 1=right)
            xanchor="center",  # Ensures proper anchoring
            font=dict(size=18),  # Optional: Set title font size
        ),
        xaxis=dict(
            title="Traces",
        ),
        yaxis=dict(
            title="Window Size",
        ),
        template="plotly_white",
    )
    fig1.update_xaxes(tickangle=90)
    return fig1

def plot_figures_segments(dist_matrix, peaks):
    labels = [f"segment{i}" for i in range(1, dist_matrix.shape[0] + 1)]
    # Create the Plotly Heatmap
    fig2 = go.Figure(
        data=go.Heatmap(
            z=dist_matrix,  # Heatmap values
            x=labels,  # X-axis labels
            y=labels,  # Y-axis labels
            colorscale="Reds",  # Color scheme
            colorbar=dict(
                title=dict(
                    text = "ldist",  # Colorbar title
                font=dict(size=18)),  # Font size for the colorbar title
                tickfont=dict(size=18)  # Font size for the colorbar ticks
            ),
            zmin=np.min(dist_matrix),  # Minimum value for color scaling
            zmax=np.max(dist_matrix),  # Maximum value for color scaling
        )
    )

    # Update layout for the figure
    fig2.update_layout(
        title=dict(
            text="Segments Comparison",
            font=dict(size=20),  # Title font size
            x=0.5  # Center the title
        ),
        xaxis=dict(
            title="",
            tickfont=dict(size=18)  # Font size for x-axis ticks
        ),
        yaxis=dict(
            title="",
            tickfont=dict(size=18)  # Font size for y-axis ticks
        ),
        template="plotly_white",
        width = 580,  # Set the width of the figure
        height = 500  # Set the height of the figure
    )

    return fig2


### previous version of the figures with mathplotlib
# def plot_figures(df, masks, n_bin, map_range, dist_matrix, peaks, w,WINDOWS):
#     every = 2
#     """Generate heatmaps and comparison plots."""
#     # Sliding Window Heatmap
#     fig1, ax1 = plt.subplots(figsize=(15, 3))
#     sns.heatmap(df, cmap="Reds", mask=np.array(masks), ax=ax1)
#     ax1.set_xticks(0.5 + np.arange(0, n_bin - 1, 3))
#     ax1.set_xticklabels(
#         [f"{round(x * (100 / n_bin))}% ({round(map_range[x], 1)})" for x in range(1, n_bin, 3)]
#     )
#     ax1.set_facecolor("gray")
#     ax1.set_title("Sliding Window Analysis")
#     ax1.set_xlabel("Traces")
#     ax1.set_ylabel("Window Size")
#     ax1.set_xticks(0.5 + np.arange(0, n_bin - 1, every))
#     ax1.set_xticklabels(
#         [str(round(x * (100 / n_bin))) + "% (" + str(round(map_range[x], 1)) + ")" for x in np.arange(1, n_bin, every)])
#     ax1.set_yticks([x+0.5 for x in range(0,len(WINDOWS))], labels=WINDOWS)
#     plt.xticks(rotation=90)
#     plt.close(fig1)
#
#     cmap = plt.cm.Reds
#     fig2 = plt.figure(figsize=(7, 7))
#     ax = sns.heatmap(dist_matrix, cmap=cmap, xticklabels=['segment' + str(i) for i in range(1, dist_matrix.shape[0] + 1)],
#                      yticklabels=['segment' + str(i) for i in range(1, dist_matrix.shape[0] + 1)])
#
#     fig2.suptitle('segments comparison', fontsize=20)
#     plt.xticks(fontsize=18)
#     plt.yticks(fontsize=18)
#     cbar = ax.collections[0].colorbar
#     cbar.ax.tick_params(labelsize=18)
#     plt.xlabel(' ', fontsize=18)
#     cbar.set_label('ldist', fontsize=18)
#     plt.close(fig2)
#
#     buf = BytesIO()
#     fig1.savefig(buf, format="png", bbox_inches = 'tight')
#     # Embed the result in the html output.
#     fig_data1 = base64.b64encode(buf.getbuffer()).decode("ascii")
#
#     buf = BytesIO()
#     fig2.savefig(buf, format="png", bbox_inches='tight')
#     # Embed the result in the html output.
#     fig_data2 = base64.b64encode(buf.getbuffer()).decode("ascii")
#     return f'data:image/png;base64,{fig_data1}', f'data:image/png;base64,{fig_data2}'
#
#
def export_logs(segments_ids):
    """Export logs for each segment."""
    case_table = pd.read_csv("output_files/out.csv")
    event_file = "output_files/out_event.csv"
    event_table = pd.read_csv(event_file)

    os.makedirs("event_logs/exported_logs", exist_ok=True)
    for idx, segment_id_set in enumerate(segments_ids, start=1):
        segment_cases = case_table.loc[segment_id_set, 'case_id']
        segment_log = event_table[event_table['case_id'].isin(segment_cases)]
        segment_log = pm4py.format_dataframe(segment_log, case_id="case_id", activity_key="activity_name",
                                             timestamp_key="timestamp")
        event_log = pm4py.convert_to_event_log(segment_log)
        pm4py.write_xes(event_log, f"event_logs/exported_logs/segment_{idx}.xes")



def apply_EMD(n_bin, w, kpi):
    """Main function to apply the analysis."""
    # if w not in WINDOWS:
    #     WINDOWS.append(w)
    #     WINDOWS.sort(reverse=True)
    WINDOWS = [w]

    log_command("binning started!")
    bins, map_range, case_table = bins_generation(kpi, n_bin)
    log_command("binning done!")

    log_command("sliding window started!")
    df, masks = sliding_window(bins, n_bin,WINDOWS)
    log_command("sliding window done!")

    redis_client.set("df", df.to_json(orient="split"))
    redis_client.set("masks", json.dumps(masks))
    redis_client.set("map_range", json.dumps(map_range))
    redis_client.set("bins",json.dumps([(b[0],b[1],list(b[2].items())) for b in bins]))
    redis_client.set("max_dist", df.iloc[0].max())

    # fig1_path, fig2_path = plot_figures(df, masks, n_bin, map_range, dist_matrix, peaks, w,WINDOWS)
    log_command("figure 1 is being generated!")
    fig1 = plot_figures_EMD(df, masks, n_bin, map_range,WINDOWS)
    log_command("figure 1 generated!")
    return fig1

def apply_segmentation(n_bin, w, signal_threshold):
    """Main function to apply the analysis."""
    df = pd.read_json(redis_client.get("df"), orient="split")


    bins = [(b[0], b[1], pd.Series(dict(b[2]))) for b in json.loads(redis_client.get("bins"))]
    log_command("segmentation started!")
    segments, segments_ids, dist_matrix, peaks = segmentation(df, bins, n_bin, w, signal_threshold)
    log_command("segmentation done!")


    redis_client.set("peaks",json.dumps(peaks))
    redis_client.set("segments_ids", json.dumps(segments_ids))



    # fig1_path, fig2_path = plot_figures(df, masks, n_bin, map_range, dist_matrix, peaks, w,WINDOWS)
    log_command("figure 2 is being generated!")
    fig2 = plot_figures_segments(dist_matrix, peaks)
    log_command("figure 2 generated!")


    map_range = json.loads(redis_client.get("map_range"))

    peak_explanations = []
    if peaks:
        for i, p in enumerate(peaks):
            if i == 0:
                peak_explanations.append(f"Segment {i + 1}: From the beginning to {round(map_range[str(peaks[0]+1)],2)}")
            else:
                peak_explanations.append(f"Segment {i + 1}: From {round(map_range[str(peaks[i - 1]+1)],2)} to {round(map_range[str(peaks[i]+1)],2)}")

        peak_explanations.append(f"Segment {len(peaks)+1}: From {round(map_range[str(peaks[-1]+1)],2)} to the end")
    else:
        peak_explanations.append("We have only one Segment!")

    return fig2,peak_explanations



import networkx as nx
from collections import Counter
import copy
from collections import defaultdict
import matplotlib.pyplot as plt
import time

EXISTENCE = "existence"
EXACTLY_ONE = "exactly_one"
INIT = "init"
RESPONDED_EXISTENCE = "responded_existence"
RESPONSE = "response"
PRECEDENCE = "precedence"
SUCCESSION = "succession"
ALTRESPONSE = "altresponse"
ALTPRECEDENCE = "altprecedence"
ALTSUCCESSION = "altsuccession"
CHAINRESPONSE = "chainresponse"
CHAINPRECEDENCE = "chainprecedence"
CHAINSUCCESSION = "chainsuccession"
ABSENCE = "absence"
COEXISTENCE = "coexistence"
NONCOEXISTENCE = "noncoexistence"
NONSUCCESSION = "nonsuccession"
NONCHAINSUCCESSION = "nonchainsuccession"

def n_edges(net, S, T, scaling = None):
    net_c = copy.deepcopy(net)
    if scaling == None:
        edges_reweight = list(nx.edge_boundary(net_c, S, T, data='weight', default=1))
    else:
        edges = list(nx.edge_boundary(net_c, S, T, data='weight', default=1))
        edges_reweight = []
        for ed in edges:
            edges_reweight.append((ed[0],ed[1], ed[2]*scaling[(ed[0], ed[1])]))
            # net_c[ed[0]][ed[1]]['weight'] = net_c[sc[0]][sc[1]]['weight'] * scaling[sc]
        # edges = edges_reweight
    return sum(weight for u, v, weight in edges_reweight if (u in S and v in T))

def drop_SE(s):
    return s-{'start','end'}

def add_SE(net, s):
    if s & set(net.successors('start')):
        s.add('start')
    if s & set(net.predecessors('end')):
        s.add('end')
    return s

def add_SS(s):
    s.add('start')
    return s

def add_EE(s):
    s.add('end')
    return s

def r_to_s(net):
    return (set(nx.descendants(net, 'start')) == (set(net.nodes) - {'start'}))


def r_from_e(net):
    return (set(nx.ancestors(net, 'end')) == (set(net.nodes) - {'end'}))

def dfg_extract(log):
    dfgs = map((lambda t: [(t[i - 1], t[i]) for i in range(1, len(t))]), log)
    return Counter([dfg for lista in dfgs for dfg in lista])

def lal(net,a):
    return net.out_degree(weight='weight')[a]

def lAl(net,A):
    return sum([net.out_degree(weight='weight')[a] for a in A])

def toggle(dic):
    dic_new = defaultdict(lambda: 1, {})
    for x in dic:
        # dic_new[x] = (1-dic[x])+1
        dic_new[x] = 1/dic[x]
    return dic_new


def cost_seq(net, A, B, in_set, out_set, sup, flow, scores):
    scores_toggle = toggle(scores)
    c1 = n_edges(net, B, A, scaling=scores_toggle)

    c2 = 0
    for x in A:
        for y in B:
            c2 += max(0, scores[(x, y)] * net.out_degree(x, weight='weight') * sup * (net.out_degree(y, weight='weight') / (
                        sum([net.out_degree(p, weight='weight') for p in B]) + sum([net.out_degree(p, weight='weight') for p in A]))) - flow[(x, y)])

    c3 = 0
    # for x in out_set:
    #     for y in in_set:
    #         c3 += max(0, scores[(x, y)] * n_edges(net, {x}, B.union({'end'}), scaling=scores) * sup * (n_edges(net, A.union({'start'}), {y}, scaling=scores) /
    #                                                                                                    (n_edges(net, A.union({'start'}), B.union({'end'}), scaling=scores))) - n_edges(net, {x}, {y}, scaling=scores))

    # c3 = 0
    # for x in out_set:
    #     for y in in_set:
    #         # c3 += max(0, scores[(x, y)] * net.out_degree(x, weight='weight') * sup * (net.out_degree(y, weight='weight') / (sum([net.out_degree(p, weight='weight') for p in B]) + sum([net.out_degree(p, weight='weight') for p in A]))) - n_edges(net, {x}, {y}, scaling=scores))
    #         if n_edges(net,out_set, in_set)==0:
    #             print('hi')
    #         c3 += max(0, scores[(x, y)] * n_edges(net,{x}, in_set) * sup * (n_edges(net,out_set, {y}) / n_edges(net,out_set, in_set)) - n_edges(net, {x}, {y}, scaling=scores))

    return c1 + c2 + c3

def fit_seq(log_var,A,B):
    count = 0
    for tr in log_var:
        for i in range(0,len(tr)-1):
            if (tr[i] in B) and (tr[i+1] in A):
                count += log_var[tr]
                break
    fit = 1-(count/sum(log_var.values()))
    return fit

def fit_exc(log_var,A,B):
    count = 0
    for tr in log_var:
        if set(tr).issubset(A) or set(tr).issubset(B):
            count += log_var[tr]
    fit = (count/sum(log_var.values()))
    return fit

def fit_loop(log_var,A,B,A_end,A_start):
    count = 0
    for tr in log_var:
        if len(tr)==0:
            continue
        if (tr[0] in B) or (tr[-1] in B):
            count += log_var[tr]
            continue
        for i in range(0,len(tr)-1):
            if (tr[i+1] in B) and (tr[i] in A):
                if (tr[i] not in A_end):
                    count += log_var[tr]
                break
            if (tr[i] in B) and (tr[i+1] in A):
                if (tr[i+1] not in A_start):
                    count += log_var[tr]
                break
    fit = 1 - (count / sum(log_var.values()))
    return fit


def cost_exc(net, A, B, scores):
    scores_toggle = toggle(scores)
    c1 = n_edges(net, A, B, scaling =scores_toggle)
    c1 += n_edges(net,B ,A, scaling =scores_toggle)
    return c1


def cost_par(net, A, B, sup, scores):
    c1 = 0
    c2 = 0
    for a in A:
        for b in B:
            c1 += max(0, scores[(a, b)] * (net.out_degree(a, weight='weight') * sup * net.out_degree(b, weight='weight')) / (
                        (sum([net.out_degree(p, weight='weight') for p in B])) + (sum([net.out_degree(p, weight='weight') for p in A]))) - n_edges(net, {a}, {b}, scaling=scores))
            c2 += max(0, scores[(b, a)] * (net.out_degree(b, weight='weight') * sup * net.out_degree(a, weight='weight')) / (
                        (sum([net.out_degree(p, weight='weight') for p in B])) + (sum([net.out_degree(p, weight='weight') for p in A]))) - n_edges(net, {b}, {a}, scaling=scores))

    return c1+c2


def cost_loop(net, A, B, sup, start_A, end_A, input_B, output_B, scores):
    scores_toggle = toggle(scores)

    flag_loop_valid = False

    if n_edges(net, B, start_A) != 0:
        if n_edges(net, end_A, B) != 0:
            flag_loop_valid = True
        else:
            return False
    else:
        return False

    BotoAs_P = n_edges(net, output_B, start_A)
    AetoBi_P = n_edges(net, end_A, input_B)
    M_P = max(BotoAs_P, AetoBi_P)



    c1 = n_edges(net, {'start'}, B, scaling=scores_toggle)
    c1 += n_edges(net, B, {'end'}, scaling=scores_toggle)

    c2 = n_edges(net, A - end_A, B, scaling= scores_toggle)

    c3 = n_edges(net, B, A - start_A, scaling=scores_toggle)

    c4 = 0
    if len(output_B) != 0:
        for a in start_A:
            for b in output_B:
                c4 += max(0, scores[(b, a)] * M_P * sup * (n_edges(net,{'start'},{a})/n_edges(net, {'start'}, start_A)) * (n_edges(net, {b}, start_A)/ n_edges(net, output_B, start_A))- n_edges(net, {b}, {a}, scaling=scores))

    c5 = 0
    if len(input_B) != 0:
        for a in end_A:
            for b in input_B:
               c5 +=  max(0, scores[(a,b)] * M_P * sup * (n_edges(net,{a}, {'end'})/n_edges(net, end_A, {'end'})) * (n_edges(net, end_A, {b})/ n_edges(net, end_A, input_B))- n_edges(net, {a}, {b}, scaling=scores))


    # if sup*M_P==0:
    #     return False
    # if (c4+c5)/(2*sup*M_P)>0.3:
    #     return False

    return c1 + c2 + c3 + c4 + c5

def visualisecpcm(cuts, ratio, size_par):
    cp = [x[2] for x in cuts]
    cm = [x[3] for x in cuts]
    tt = [str(x[1])+", "+str(x[0]) for x in cuts]
    diff = [x[2] - ratio * size_par * x[3] for x in cuts]
    color_fit = [x[5] for x in cuts]
    min_value = min(diff)
    min_index = diff.index(min_value)
    edge = [20]*len(diff)
    edge[min_index]= 100

    fig, ax = plt.subplots()
    s = ax.scatter(cp, cm, c=color_fit, cmap='inferno',s=edge)
    ax.set_xlabel(r'cp', fontsize=15)
    ax.set_ylabel(r'cm', fontsize=15)
    fig.colorbar(s, ax=ax)



    from matplotlib.widgets import Cursor
    # Defining the cursor
    cursor = Cursor(ax, horizOn=True, vertOn=True, useblit=True,
                    color='r', linewidth=1)

    # cursor grid lines
    lnx = plt.plot([60, 60], [0, 1.5], color='black', linewidth=0.3)
    lny = plt.plot([0, 100], [1.5, 1.5], color='black', linewidth=0.3)
    lnx[0].set_linestyle('--')
    lny[0].set_linestyle('None')
    # annotation
    annot = ax.annotate("", xy=(0, 0), xytext=(5, 5), textcoords="offset points")
    annot.set_visible(False)
    # xy limits
    plt.xlim(min(cp) * 0.95, max(cp) * 1.05)
    plt.ylim(min(cm) * 0.95, max(cm) * 1.05)

    def hover(event):
        # check if event was in the axis
        if event.inaxes == ax:
            cont, ind = s.contains(event)
            if cont:
                # change annotation position
                annot.xy = (event.xdata, event.ydata)
                print((event.xdata, event.ydata))
                print("{}".format(', '.join([tt[n] for n in ind["ind"]])))
                # write the name of every point contained in the event
                annot.set_text("{}".format('\n '.join([tt[n] for n in ind["ind"]])))
                annot.set_visible(True)
                fig.canvas.draw()
            else:
                annot.set_visible(False)
        # else:
        #     lnx[0].set_visible(False)
        #     lny[0].set_visible(False)

    fig.canvas.mpl_connect("motion_notify_event", hover)
    plt.show()


def check_base_case(self, logP, logM, sup_thr, ratio, size_par):
    activitiesP = set(a for x in logP.keys() for a in x)

    if len(activitiesP) <= 1:
        base_check = True
        counter = logP[()]
        counterM = logM[()]
        len_logP = sum(logP.values())
        acc_contP = sum([len(x) * logP[x] for x in logP])
        len_logM = sum(logM.values())
        acc_contM = sum([len(x) * logM[x] for x in logM])

        # empty check
        if (counter == len_logP) or (len_logP == 0):
            self.detected_cut = 'empty_log'
            cut = ('none', 'empty_log', 'none', 'none')
        else:
            # xor check
            cost_single_exc = max(0, sup_thr * len_logP - counter) - ratio * size_par * max(0,sup_thr * len_logM - counterM)
            if (counter > (sup_thr / 2) * len_logP) and (cost_single_exc <= 0):
            # if (cost_single_exc <= 0):
                cut = (({activitiesP.pop()}, set()), 'exc', 'none', 'none')
            else:
                # loop check
                del logP[()]
                if acc_contP > 0:
                    p_prime_Lp = (len_logP - counter) / ((len_logP - counter) + acc_contP)
                else:
                    p_prime_Lp = 'nd'

                if acc_contM > 0:
                    p_prime_Lm = (len_logM - counterM) / ((len_logM - counterM) + acc_contM)
                else:
                    p_prime_Lm = 'nd'

                if p_prime_Lm != 'nd':
                    cost_single_loop = max(0, sup_thr/2 - abs(p_prime_Lp - 0.5)) - ratio * size_par * max(0,sup_thr/2 - abs(p_prime_Lm - 0.5))
                else:
                    cost_single_loop = max(0, sup_thr/2 - ratio * size_par * abs(p_prime_Lp - 0.5))

                if (abs(p_prime_Lp - 0.5) > sup_thr / 2) and (cost_single_loop <= 0):
                # if (cost_single_loop <= 0):
                    cut = (({activitiesP.pop()}, set()), 'loop1', 'none', 'none')
                else:
                    # single activity
                    self.detected_cut = 'single_activity'
                    cut = ('none', 'single_activity', 'none', 'none')
    else:
        base_check = False
        cut = "not_base"

    return base_check, cut




def get_all_connected_subgraphs(G):
    """Get all connected subgraphs by a recursive procedure"""
    # G = N.to_undirected(reciprocal=False, as_view=False)
    con_comp = [c for c in sorted(nx.strongly_connected_components(G), key=len, reverse=True)]

    def recursive_local_expand(node_set, possible, excluded, results, max_size):
        """
        Recursive function to add an extra node to the subgraph being formed
        """
        results.append(node_set)
        if len(node_set) == max_size:
            return
        for j in possible - excluded:
            new_node_set = node_set | {j}
            excluded = excluded | {j}
            new_possible = (possible | set(G.neighbors(j))) - excluded
            recursive_local_expand(new_node_set, new_possible, excluded, results, max_size)

    results = []
    for cc in con_comp:
        max_size = len(cc)

        excluded = set()
        for i in G:
            excluded.add(i)
            recursive_local_expand({i}, set(G.neighbors(i)) - excluded, excluded, results, max_size)

    results.sort(key=len)

    return results



def find_possible_partitions3(net,rules,sa,ea):

    def adj(node_set, net):
        adj_set = set()
        for node in node_set:
            adj_set = adj_set.union(set(net.neighbors(node)))
        return adj_set

    activity_list = set(net.nodes)-{'start','end'}
    queue = [(set(), {'start'})]
    # queue.append()
    visited = []
    valid = []
    # reserve_list =[]
    st_net = {x:net.succ['start'][x]['weight'] for x in net.succ['start'] if x!='end'}
    en_net = {x: net.pred['end'][x]['weight'] for x in net.pred['end'] if x!='end'}
    while len(queue) != 0:
        current = queue.pop()
        for x in current[1]:
            new_state = current[0].union({x})
            if x in sa:
                new_state.add('start')
            if x in ea:
                new_state.add('end')

            # new_state = add_SE(net, new_state)
            if new_state not in visited:
                new_adj = current[1].union(adj({x},net)) - new_state
                visited.append(new_state)


                B = activity_list - new_state
                if (len(B) == 0) or (len(B) == len(activity_list)):
                    queue.append((new_state, new_adj))
                    if (len(B) == 0) and (n_edges(net, set(en_net.keys()), set(st_net.keys()), scaling=None)>0):
                        na, block, n_conflicts = is_allowed(new_state, B, rules, st_net, en_net)
                        possible_cuts = {'loop_tau'} - na
                        if len(possible_cuts) > 0:
                            valid.append((new_state, B, possible_cuts))


                    continue
                B = add_SE(net, B)
                netA = net.subgraph(new_state)
                netB = net.subgraph(B)
                # BB = net.subgraph(B)
                # na = {}
                block = False

                # 'start' ~> netB
                if 'start' in netB:
                    startB = set(netB.nodes) - set(nx.descendants(netB, 'start')) - {'start','end'}
                else:
                    startB = set(netB.nodes)
                # 'end' ~> netA
                if 'end' in netA:
                    endA = set(netA.nodes) - set(nx.ancestors(netA, 'end')) - {'start','end'}
                else:
                    endA = set(netA.nodes)
                # 'end' ~> netB
                if 'end' in netB:
                    endB = set(netB.nodes) - set(nx.ancestors(netB, 'end')) - {'start', 'end'}
                else:
                    endB = set(netB.nodes)

                na, block, n_conflicts = is_allowed(new_state, B, rules,st_net,en_net)
                possible_cuts = set()
                if len(endB)==0:
                    possible_cuts.add("seq")
                    if len(endA)==0:
                        possible_cuts.add("loop")
                        if len(startB)==0 and (B not in visited):
                            possible_cuts.add("exc")
                            possible_cuts.add("par")
                elif len(endA)==0:
                    possible_cuts.add("loop")
                # r_p = possible_cuts
                possible_cuts = possible_cuts - na
                if len(possible_cuts)>0:
                    valid.append((new_state, B, possible_cuts))
                # else:
                #     reserve_list.append((n_conflicts,(new_state, B, r_p)))
                if block==False:
                    queue.append((new_state, new_adj))
    # if len(valid)==0:
    #     m_conf = min([n[0] for n in reserve_list])
    #     more = [c[1] for c in reserve_list if c[0]==m_conf]
    #     valid = valid + more
    #     print(f'conf:{m_conf}')
    return valid




def is_allowed(S1,S2,rules,st_net,en_net):
    exclude = []
    n_conf = 0
    block = False

    # if (S2-{'start', 'end'}) == {'A_Incomplete'}:
    #     print('hh')

    # o_S1 = sorted([rules['order'].index(x) for x in S1-{'start', 'end'} if x in rules['order']])
    # o_S2 = sorted([rules['order'].index(x) for x in S2-{'start', 'end'} if x in rules['order']])
    # if len(o_S1)>0 and len(o_S2)>0 and o_S1[-1]>o_S2[0]:
    #     exclude.append('seq')
    #     block=True



    s_count = 0
    e_count = 0
    for n in S2-{'start','end'}:
        if n in st_net:
            s_count += st_net[n]
        if n in en_net:
            e_count += en_net[n]
    if s_count>= 0.05 * sum(st_net.values()) or e_count>= 0.05 * sum(en_net.values()):
        n_conf+=1
        exclude.append('loop')

    EXISTENCE_acts = set([r for r in rules[EXISTENCE]])
    EXACTLY_ONE_acts = set([r for r in rules[EXACTLY_ONE]])

    if S1.intersection(EXACTLY_ONE_acts) == S1:
        n_conf += 1
        exclude.append('loop')

    if S2.intersection(EXISTENCE_acts) == S2:
        n_conf += 1
        exclude.append('loop')


    for r in rules[NONSUCCESSION]:
        if r[0] in S1 and r[1] in S2:
            n_conf += 1
            exclude.append('seq')
            # exclude.append('loop')
            exclude.append('par')
            # block = True
        elif r[0] in S2 and r[1] in S1:
            n_conf += 1
            exclude.append('par')
            exclude.append('loop')
            # block = True
        elif (r[0] in S1 and r[1] in S1) or (r[0] in S2 and r[1] in S2):
            exclude.append('loop')
            # exclude.append('loop_tau')


    for r in rules[RESPONDED_EXISTENCE]:
        if r[0] in S1 and r[1] in S2:
            n_conf += 1
            exclude.append('exc')
            # exclude.append('loop')
        elif r[0] in S2 and r[1] in S1:
            n_conf += 1
            exclude.append('exc')


    for r in rules[NONCOEXISTENCE]:
        if r[0] in S1 and r[1] in S2:
            exclude.append('par')
            exclude.append('seq')
            exclude.append('loop')
        elif r[0] in S2 and r[1] in S1:
            exclude.append('par')
            exclude.append('seq')
            exclude.append('loop')
        elif (r[0] in S1 and r[1] in S1) or (r[0] in S2 and r[1] in S2):
            # exclude.append('loop_tau')
            print("ss")

        # if (r[0] in S1 and r[1] in S1):
        #     block = True

    for r in rules[RESPONSE]:
        if (r[1], r[0]) not in rules[PRECEDENCE]:
            if r[0] in S1 and r[1] in S2:
                n_conf += 1
                exclude.append('exc')
                exclude.append('loop')
                exclude.append('par')
            if r[0] in S2 and r[1] in S1:
                n_conf += 1
                exclude.append('exc')
                exclude.append('seq')
                exclude.append('par')
                block = True
        else:
            if r[0] in S1 and r[1] in S2:
                n_conf += 1
                exclude.append('exc')
                exclude.append('loop')
                exclude.append('par')

            if r[0] in S2 and r[1] in S1:
                n_conf += 1
                exclude.append('exc')
                exclude.append('loop')
                exclude.append('par')



    for r in rules[PRECEDENCE]:
        if (r[1], r[0]) not in rules[RESPONSE]:
            if r[0] in S1 and r[1] in S2:
                n_conf += 1
                exclude.append('par')
                exclude.append('exc')
            if r[0] in S2 and r[1] in S1:
                n_conf += 1
                exclude.append('exc')
                exclude.append('seq')
                exclude.append('par')
                exclude.append('loop')
                block = True
        else:
            if r[0] in S1 and r[1] in S2:
                n_conf += 1
                exclude.append('par')
                exclude.append('exc')
                exclude.append('seq')
                exclude.append('loop')
            if r[0] in S2 and r[1] in S1:
                n_conf += 1
                exclude.append('exc')
                exclude.append('par')
                exclude.append('loop')



    return set(exclude), block, n_conf


def max_flow_graph(net):
    flow_graph = {}
    for x in net.nodes:
        for y in net.nodes:
            if (x != y):
                flow_graph[(x, y)] = nx.algorithms.flow.maximum_flow(net, x, y, capacity='weight')[0]
    return flow_graph



def noise_filtering(dfg0, nt):
    dfg = copy.deepcopy(dfg0)
    log_size = sum([dfg[x] for x in dfg if x[0] == 'start'])
    noisy_edges = sorted([(x,dfg[x]) for x in dfg if (dfg[x]/log_size) < nt], key=lambda z:z[1])
    net = generate_nx_graph_from_dfg(dfg0)
    for ne in noisy_edges:
        net_copy = copy.deepcopy(net)
        nodes_set = set(net_copy.nodes)
        net_copy.remove_edge(ne[0][0],ne[0][1])
        if (set(nx.ancestors(net_copy, 'end')) == nodes_set-{'end'}):
            if(set(nx.descendants(net_copy, 'start')) == nodes_set-{'start'}):
                del dfg[ne[0]]
                net = net_copy
    return dfg

def generate_nx_graph_from_dfg(dfg):
    dfg_acts = set()
    for x in dfg:
        dfg_acts.add(x[0])
        dfg_acts.add(x[1])
    G = nx.DiGraph()
    for act in dfg_acts:
        G.add_node(act)
    for edge in dfg:
        G.add_edge(edge[0], edge[1])
    return G


from enum import Enum, auto
import itertools
import sys
import time
import heapq
import copy

from typing import Dict, List, Optional, Tuple, Union
from dataclasses import dataclass
from collections import defaultdict
from itertools import count

import numba as nb
import numpy as np
import numpy.typing as npt
import pm4py
from pm4py.objects.log.obj import EventLog
from pm4py.objects.petri_net.obj import PetriNet, Marking
from pm4py.objects.petri_net.semantics import ClassicSemantics

import ot

from prolysis.analysis.python_emsc.util import time_it


VariantLog = List[Tuple[Tuple[str], int]]

WFNet = Tuple[PetriNet, Marking, Marking]
PNTransition = PetriNet.Transition

class EMSCType(Enum):
    EMSC = auto()
    PEMSC = auto()


@dataclass
class StochasticPetriNet:
    wf_net: WFNet
    weights: Dict[PNTransition, float]


# EventLog implicitly defines a stochastic language
StochasticLanguageObject = Union[EventLog, StochasticPetriNet]


def get_variant_log(log: EventLog) -> VariantLog:
    variants = pm4py.get_variants_as_tuples(log)
    #print(variants)
    #return [(variant, len(traces)) for variant, traces in variants.items()]
    return [(variant, nbr_traces) for variant, nbr_traces in variants.items()]


def log_to_stochastic_language(log: EventLog) -> Dict[Tuple[str], float]:
    variant_log = get_variant_log(log)
    trace_count = sum(freq for _, freq in variant_log)
    return {trace: freq / trace_count for trace, freq in variant_log}


def view_stochastic_petri_net(stochastic_pn: StochasticPetriNet):
    # TODO it is quite painful that PM4Py returns a GraphViz object that cannot be further edited
    #   We work around that by adding the weight to the transition label. Taus become TAU:(weight)
    #   In the future, use my own visualization utilities
    net, im, fm = stochastic_pn.wf_net
    weights = {transition.name: weight for transition, weight in stochastic_pn.weights.items()}

    new_net = copy.deepcopy(net)
    for transition in new_net.transitions:
        if transition.label is None:
            transition.label = f'TAU:{weights[transition.name]}'
        else:
            transition.label = f'{transition.label}:{weights[transition.name]}'

    pm4py.view_petri_net(new_net, im, fm)


def sample_most_likely_net_paths(
        stochastic_net: StochasticPetriNet, prob_mass: Optional[float] = None,
        max_num_paths: Optional[int] = None, max_total_time_ms: Optional[int] = None
) -> List[Tuple[List[PNTransition], float]]:
    """
    Samples the most likely paths from the Petri net until one of the stop criteria is reached.
    If a parameter is None, then it's not considered.
    """
    assert prob_mass is not None or max_num_paths is not None or max_total_time_ms is not None, \
        "At least one of the stopping criteria must be specified."

    if prob_mass is None:
        prob_mass = 1.0
    if max_num_paths is None:
        max_num_paths = sys.maxsize
    if max_total_time_ms is None:
        max_total_time_ms = sys.maxsize
    max_total_time_ns = max_total_time_ms * 1_000_000

    net, im, fm = stochastic_net.wf_net
    weights = stochastic_net.weights

    paths = []
    sampled_mass = 0.0
    semantics = ClassicSemantics()
    marking_transition_cache: Dict[Marking, List[Tuple[PNTransition, Marking, float]]] = {}

    # heapq is one of the few situations where C++ is easier to use than Python
    # We use tuples because the overhead of class
    to_explore = []
    # To break ties in the queue
    id_counter = itertools.count(0)

    # path_id -> (incoming_transition, prev_path)
    path_transition_function = {}

    def _push_heap(_prob, _prev_path_id, _input_transition, _marking):
        path_id = next(id_counter)
        # Keep the heap compact by storing other data outside
        path_transition_function[path_id] = _prev_path_id, _input_transition
        # Must invert the probability to get a max-heap
        heapq.heappush(to_explore, (-_prob, path_id, _marking))

    start_time = time.time_ns()

    def _should_terminate():
        return sampled_mass >= prob_mass or len(paths) >= max_num_paths \
            or (time.time_ns() - start_time) > max_total_time_ns

    _push_heap(1.0, None, None, im)
    it_counter = itertools.count(0)
    while to_explore:
        it_count = next(it_counter)
        if it_count % 100 == 0 and _should_terminate():
            break
        #if it_count % 1000 == 0:
        #    print(sampled_mass, len(paths), len(to_explore))

        cur_prob, cur_path_id, cur_marking = heapq.heappop(to_explore)
        cur_prob = -cur_prob

        if cur_marking not in marking_transition_cache:
            enabled_transitions = semantics.enabled_transitions(net, cur_marking)
            weight_sum = sum(weights[transition] for transition in enabled_transitions)
            marking_transition_cache[cur_marking] = [
                (transition, semantics.weak_execute(transition, net, cur_marking),
                 weights[transition] / weight_sum)
                for transition in enabled_transitions
            ]

        transition_function = marking_transition_cache[cur_marking]
        if len(transition_function) == 0:
            # Reconstruct path
            path = []
            prev_path_id, prev_transition = path_transition_function[cur_path_id]
            while prev_transition is not None:
                path.append(prev_transition)
                prev_path_id, prev_transition = path_transition_function[prev_path_id]
            path.reverse()
            paths.append((path, cur_prob))
            sampled_mass += cur_prob
            continue

        for transition, tgt_marking, transition_prob in transition_function:
            tgt_prob = cur_prob * transition_prob
            assert tgt_prob != 0.0
            _push_heap(tgt_prob, cur_path_id, transition, tgt_marking)

    distinct_paths = {''.join(transition.name for transition in path) for path, _ in paths}
    assert len(distinct_paths) == len(paths), \
        "Cannot sample the same path twice. This indicates an error in the logic"

    path_probs = [path_prob for _, path_prob in paths]
    assert all(prev_prob >= next_prob for prev_prob, next_prob in zip(path_probs, path_probs[1:])), \
        "Paths should be generated from most to least likely"

    return paths


def path_language_to_trace_language(path_language: List[Tuple[List[PNTransition], float]]) \
        -> Dict[Tuple[str], float]:
    # A path language is a sampling over paths. The language maps the path to its labels
    language = defaultdict(lambda: 0.0)
    for path, path_prob in path_language:
        trace = tuple(transition.label for transition in path if transition.label is not None)
        language[trace] += path_prob

    #print(f"PATHS SIZE {len(path_language)} \t LANGUAGE SIZE {len(language)}")
    return dict(language)


@time_it
def compare_languages_levenshtein(lhs: Dict[Tuple[str], float], rhs: Dict[Tuple[str], float], emsc_type: EMSCType=EMSCType.EMSC) \
        -> float:
    # "Assign" traces to IDs that can be used as matrix indexes
    lhs = [*lhs.items()]
    rhs = [*rhs.items()]

    # Compute distance matrix
    #print(f'BUILDING MATRIX')
    if len(lhs) == 0 or len(rhs) == 0:
        raise Exception(f'Cannot compute distance matrix for size {len(lhs)} x {len(rhs)}!')
    dist_matrix = build_distance_matrix(lhs, rhs)

    #print(f'SOLVING SYSTEM')
    prob_src_dist = np.array([src_prob for _, src_prob in lhs])
    # Target distribution is SPN
    prob_tgt_dist = np.array([tgt_prob for _, tgt_prob in rhs])
    # Check if fully unrolled
    diff_mass = np.sum(prob_src_dist) - np.sum(prob_tgt_dist)
    assert diff_mass > -0.0001

    # Add "residual" target, cost function depends on EMSC type
    if diff_mass > 0.000001:
        prob_tgt_dist = np.append(prob_tgt_dist, diff_mass)
        if emsc_type == EMSCType.EMSC:
            # Extend by column that contains minimum distance for each source
            dist_matrix = np.concatenate((dist_matrix, np.min(dist_matrix, axis=1)[:, np.newaxis]), axis=1)
            pass
        elif emsc_type == EMSCType.PEMSC:
            # Extend by column of ones
            dist_matrix = np.concatenate((dist_matrix, np.ones((len(prob_src_dist), 1))), axis=1)
        else:
            raise Exception('Unknown EMSC Type')

    # Build and solve the EMD problem (old lemon wrapper)
    #start_time_wasserstein = time.time_ns()
    #emd_cost = solve_emd(prob_src_dist, prob_tgt_dist, dist_matrix)
    #time_wasserstein = time.time_ns() - start_time_wasserstein

    # OT library (also wrapper to lemon but seems to be more stable)
    #start_time_ot = time.time_ns()
    emd_cost_ot = solve_emd_ot(prob_src_dist, prob_tgt_dist, dist_matrix)

    return emd_cost_ot
    #time_ot = time.time_ns() - start_time_ot

    #if abs(emd_cost - emd_cost_ot) > 0.005:
    #    print(f'PROBLEM EMD MISMATCH: {emd_cost} | {emd_cost_ot} | mass diff {diff_mass}')
    #assert abs(emd_cost - emd_cost_ot) < 0.005

    #if (emd_cost < 0) and (emd_cost_ot < 0):
    #    return (-1, time_wasserstein, time_ot)
    #elif emd_cost_ot >= 0:
    #    return (1.0 - (emd_cost_ot - diff_mass), time_wasserstein, time_ot)
    #else:
    #    return (1.0 - (emd_cost - diff_mass), time_wasserstein, time_ot)



@nb.njit('uint32(uint8[::1], uint8[::1], uint32[::1])', inline='always', cache=True)
def _compiled_distance_lvs_pref_suff_aware(s0: np.array, s1: np.array, previous_row_tmp: np.array):
    start_index = 0
    source_end = len(s0)
    target_end = len(s1)

    while start_index < source_end and start_index < target_end and s0[start_index] == s1[start_index]:
        start_index += 1


    while start_index < source_end and start_index < target_end and s0[source_end - 1] == s1[target_end - 1]:
        source_end -= 1
        target_end -= 1

    if start_index == source_end: 
        return target_end - start_index
    elif start_index == target_end:
        return source_end - start_index


    for i in range(start_index, target_end):
        previous_row_tmp[i] = i + 1 - start_index

    for i in range(start_index, source_end):
        prev_col = i + 1 - start_index
        prev_diag = i - start_index

        for j in range(start_index, target_end):
            prev_row = previous_row_tmp[j]

            insertOrDelete = (prev_col if prev_col < prev_row else prev_row) + 1
            edit = prev_diag + (0 if (s0[i] == s1[j]) else 1)

            prev_diag = prev_row
            prev_col = edit if edit < insertOrDelete else insertOrDelete
            previous_row_tmp[j] = prev_col

    return prev_col


@nb.njit('uint32(uint8[::1], uint8[::1], uint32[::1])', inline='always', cache=True)
def _compiled_distance(s0: np.array, s1: np.array, previous_row_tmp: np.array):
    for i in range(len(s1)):
        previous_row_tmp[i] = i + 1
    
    for i in range(len(s0)):
        prev_col = i + 1
        prev_diag = i

        for j in range(len(s1)):
            prev_row = previous_row_tmp[j]

            insertOrDelete = (prev_col if prev_col < prev_row else prev_row) + 1
            edit = prev_diag + (0 if (s0[i] == s1[j]) else 1)

            prev_diag = prev_row
            prev_col = edit if edit < insertOrDelete else insertOrDelete
            previous_row_tmp[j] = prev_col

    return prev_col


@nb.njit("uint32(uint8[::1], uint8[::1], uint32[::1], uint32[::1])", inline='always', cache=True)
def _compiled_distance_two_rows(s0: np.array, s1: np.array, v0: np.array, v1: np.array):
    # This is a copy of strsimpy's implementation but optimized
    for idx in range(len(s1) + 1):
        v0[idx] = idx
        v1[idx] = 0


    for i in range(len(s0)):
        v1[0] = i + 1
        a = v1[0]
        for j in range(len(s1)):
            cost = 1
            if s0[i] == s1[j]:
                cost = 0
            a = v1[j] + 1
            b = v0[j + 1] + 1
            c = v0[j] + cost
            min_el = a if a <= b else b
            v1[j + 1] = min_el if min_el <= c else c
            # ===========
            # # Even more optimized
            # b = v0[j + 1]
            # min_el = a if a <= b else b
            # min_el += 1
            # if s0_i == s1[j]:
            #     c = v0[j]
            #     min_el = c if c < min_el else min_el
            # v1[j + 1] = min_el
            # a = min_el

        v0, v1 = v1, v0

    return v0[len(s1)]  # Example: Return the final cost (edit distance)


@nb.njit('float64(uint8[::1], uint8[::1], uint32[::1])', inline='always', cache=True)
def normalized_lev_distance(s0: np.array, s1: np.array, np_previous_row_tmp: np.array):
    if len(s0) == 0 and len(s1) == 0:
        return 0.0

    lev_distance = _compiled_distance(s0, s1, np_previous_row_tmp)
    #lev_distance = _compiled_distance_lvs_pref_suff_aware(s0, s1, np_previous_row_tmp)

    return lev_distance / max(1, max(len(s0), len(s1)))


@nb.njit('float64(uint8[::1], uint8[::1], uint32[::1], uint32[::1])', inline='always', cache=True)
def normalized_lev_distance_2(s0: np.array, s1: np.array, v0: np.array, v1):
    if len(s0) == 0 and len(s1) == 0:
        return 0.0

    lev_distance = _compiled_distance_two_rows(s0, s1, v0, v1)

    return lev_distance / max(1, max(len(s0), len(s1)))

def build_distance_matrix(
        src_dist: List[Tuple[Tuple[str], float]],
        tgt_dist: List[Tuple[Tuple[str], float]],
) -> np.ndarray:
    # Map labels to np_arrays
    counter = count(0)
    label_map = defaultdict(lambda: next(counter))

    src_dist = [np.array([label_map[label] for label in trace], dtype=np.uint8) for trace, _ in src_dist]
    tgt_dist = [np.array([label_map[label] for label in trace], dtype=np.uint8) for trace, _ in tgt_dist]

    # To numba
    src_dist_numba = nb.typed.List(src_dist)
    tgt_dist_numba = nb.typed.List(tgt_dist)

    d = fill_distance_matrix(src_dist_numba, tgt_dist_numba, normalized_lev_distance)
    #d = fill_distance_matrix_two_row(src_dist_numba, tgt_dist_numba, normalized_lev_distance_2)
    return d


@time_it
@nb.njit(parallel=True)
def fill_distance_matrix(
        src_language: nb.typed.List[np.array],
        tgt_language: nb.typed.List[np.array],
        distance_function
):
    # We do this ugly implementation because we want Numba to effectively compile it
    dist_matrix = np.empty(shape=(len(src_language), len(tgt_language)), dtype=np.float32)

    for src_id in nb.prange(len(src_language)):
        src_trace = src_language[src_id]
        np_previous_row_tmp = np.zeros(len(src_trace), dtype=np.uint32)
        for tgt_id, tgt_trace in enumerate(tgt_language):
            dist_matrix[src_id, tgt_id] = distance_function(tgt_trace, src_trace, np_previous_row_tmp)

    return dist_matrix


@time_it
@nb.njit(parallel=True)
def fill_distance_matrix_two_row(
        src_language: nb.typed.List[np.array],
        tgt_language: nb.typed.List[np.array],
        distance_function
):
    # We do this ugly implementation because we want Numba to effectively compile it
    dist_matrix = np.empty(shape=(len(src_language), len(tgt_language)), dtype=np.float32)

    for src_id in nb.prange(len(src_language)):
        src_trace = src_language[src_id]
        # Will be filled inside
        v0 = np.zeros(len(src_trace) + 1, dtype=np.uint32)
        v1 = np.zeros(len(src_trace) + 1, dtype=np.uint32)
        for tgt_id, tgt_trace in enumerate(tgt_language):
            dist_matrix[src_id, tgt_id] = distance_function(tgt_trace, src_trace, v0, v1)

    return dist_matrix


@time_it
def solve_emd_ot(
        prob_src_dist: npt.NDArray[np.float32],
        prob_tgt_dist: npt.NDArray[np.float32],
        distance_matrix: npt.NDArray[np.float32]
):
    try:
        # With transport matrix
        #(gamma, log) = ot.emd(prob_src_dist, prob_tgt_dist, distance_matrix, log=True, numItermax=500_000)
        # return log['cost']
        # Without transport matrix
        emd = ot.emd2(prob_src_dist, prob_tgt_dist, distance_matrix, log=False, numItermax=500_000)
        return emd
    except Exception as e:
        print(e)
        return -1



import prolysis.discovery.split_functions.split as split
from prolysis.discovery.candidate_search.search import find_possible_partitions
from prolysis.discovery.base_case.check import check_base_case
from prolysis.discovery.cut_quality.cost_functions import cost_functions
from prolysis.util.functions import generate_nx_graph_from_log, generate_nx_indirect_graph_from_log, read_append_write_json
from prolysis.discovery.cut_quality.cost_functions.cost_functions import overal_cost

class SubtreePlain(object):
    def __init__(self, logp,logm, rec_depth, noise_threshold= None,
                   parameters=None, sup= None, ratio = None, size_par = None, rules = None):

        nt = 0.000 * logp.total()
        self.rec_depth = rec_depth
        self.noise_threshold = noise_threshold
        self.log = logp
        self.netP = generate_nx_graph_from_log(self.log,nt)
        if self.log.total()==0:
            self.activitiesP = set()
            self.start_activitiesP = {}
            self.end_activitiesP = {}
        else:
            self.activitiesP = set(self.netP.nodes()) - {'start', 'end'}
            self.start_activities = {x:self.netP.get_edge_data('start', x)['weight'] for x in self.netP.successors('start') if x!='end'}
            self.end_activities = {x: self.netP.get_edge_data(x, 'end')['weight'] for x in self.netP.predecessors('end') if x!='start'}
        self.logM = logm
        self.netM = generate_nx_graph_from_log(self.logM,nt)
        if self.logM.total()==0:
            self.activitiesM = set()
            self.start_activitiesM = {}
            self.end_activitiesM = {}
            self.size_adj = 1
        else:
            self.activitiesM = set(self.netM.nodes()) - {'start', 'end'}
            self.start_activitiesM = {x: self.netM.get_edge_data('start', x)['weight'] for x in
                                      self.netM.successors('start') if x != 'end'}
            self.end_activitiesM = {x: self.netM.get_edge_data(x, 'end')['weight'] for x in
                                    self.netM.predecessors('end') if x != 'start'}
            self.size_adj = self.log.total()/self.logM.total()
        self.original_log = logp
        self.parameters = parameters
        self.detected_cut = None
        self.children = []
        self.detect_cut(parameters=parameters, sup= sup, ratio = ratio, size_par = size_par, rules = rules)


    def detect_cut(self, parameters=None, sup= None, ratio = None, size_par = None, rules = None):

        if parameters is None:
            parameters = {}

        # check base cases:
        isbase, cut = check_base_case(self.netP, self.netM,rules , sup, ratio, self)

        if isbase==False:
            fP = generate_nx_indirect_graph_from_log(self.log)
            fM = generate_nx_indirect_graph_from_log(self.logM)

            possible_partitions,reserve = find_possible_partitions(self.netP,rules,self.start_activities, self.end_activities,fP)


            cut = []

            ratio_backup = ratio

            for pp in possible_partitions:
                A = pp[0] - {'start', 'end'}
                B = pp[1] - {'start', 'end'}

                start_A_P = set(self.start_activities.keys()) & A
                end_A_P = set(self.end_activities.keys()) & A
                input_B_P = set([x[1] for x in self.netP.edges if ((x[0] not in B) and (x[1] in B))])
                output_B_P = set([x[0] for x in self.netP.edges if ((x[0] in B) and (x[1] not in B))])

                start_A_M = set(self.start_activitiesM.keys()) & A
                end_A_M = set(self.end_activitiesM.keys()) & A
                input_B_M = set([x[1] for x in self.netM.edges if ((x[0] not in B) and (x[1] in B))])
                output_B_M = set([x[0] for x in self.netM.edges if ((x[0] in B) and (x[1] not in B))])

                type = pp[2]
                if self.logM.total()==0:
                    ratio = 0
                else:
                    ratio = ratio_backup

                #####################################################################
                # seq check
                if type=="seq":
                    cost_seq_P_dict = cost_functions.cost_seq(self.netP, A, B, sup, fP)
                    cost_seq_P = sum(x['missing']+x['deviating'] for x in cost_seq_P_dict.values())
                    cost_seq_M_dict = cost_functions.cost_seq(self.netM, A.intersection(self.activitiesM), B.intersection(self.activitiesM), sup, fM)
                    cost_seq_M = sum(x['missing']+x['deviating'] for x in cost_seq_M_dict.values())
                    cut.append(((A, B), 'seq', cost_seq_P, cost_seq_M, overal_cost(cost_seq_P,cost_seq_M,ratio,self.size_adj),cost_seq_P_dict,cost_seq_M_dict))
                #####################################################################
                # xor check
                if type=="exc":
                    cost_exc_P_dict = cost_functions.cost_exc(self.netP, A, B)
                    cost_exc_P = sum(x['missing']+x['deviating'] for x in cost_exc_P_dict.values())
                    cost_exc_M_dict = cost_functions.cost_exc(self.netM, A.intersection(self.activitiesM), B.intersection(self.activitiesM))
                    cost_exc_M = sum(x['missing']+x['deviating'] for x in cost_exc_M_dict.values())
                    cut.append(((A, B), 'exc', cost_exc_P, cost_exc_M, overal_cost(cost_exc_P,cost_exc_M,ratio,self.size_adj),cost_exc_P_dict,cost_exc_M_dict))
                #####################################################################
                # xor-tau check
                if type=="exc_tau":
                    cost_exc_tau_P_dict = cost_functions.cost_exc_tau(self.netP, sup)
                    cost_exc_tau_P = sum(x['missing']+x['deviating'] for x in cost_exc_tau_P_dict.values())
                    cost_exc_tau_M_dict = cost_functions.cost_exc_tau(self.netM, sup)
                    cost_exc_tau_M = sum(x['missing']+x['deviating'] for x in cost_exc_tau_M_dict.values())
                    cut.append(((A.union(B), set()), 'exc_tau',cost_exc_tau_P , cost_exc_tau_M, overal_cost(cost_exc_tau_P,cost_exc_tau_M,ratio,self.size_adj),cost_exc_tau_P_dict,cost_exc_tau_M_dict))
                #####################################################################
                # parallel check
                if type=="par":
                    cost_par_P_dict = cost_functions.cost_par(self.netP, A, B,sup)
                    cost_par_P = sum(x['missing']+x['deviating'] for x in cost_par_P_dict.values())
                    cost_par_M_dict = cost_functions.cost_par(self.netM, A.intersection(self.activitiesM), B.intersection(self.activitiesM), sup)
                    cost_par_M = sum(x['missing']+x['deviating'] for x in cost_par_M_dict.values())
                    cut.append(((A, B), 'par', cost_par_P, cost_par_M, overal_cost(cost_par_P,cost_par_M,ratio,self.size_adj),cost_par_P_dict,cost_par_M_dict))
                #####################################################################
                # loop check
                if type=="loop":
                    cost_loop_P_dict = cost_functions.cost_loop(self.netP, A, B, sup, start_A_P, end_A_P, input_B_P,output_B_P,self.start_activities,self.end_activities)
                    cost_loop_P = sum(x['missing']+x['deviating'] for x in cost_loop_P_dict.values())
                    # if cost_loop_P==0:
                    #     print('wait')
                    cost_loop_M_dict = cost_functions.cost_loop(self.netM, A, B, sup, start_A_M, end_A_M, input_B_M,output_B_M,self.start_activitiesM,self.end_activitiesM)
                    cost_loop_M = sum(x['missing']+x['deviating'] for x in cost_loop_M_dict.values())
                    cut.append(((A, B), 'loop', cost_loop_P, cost_loop_M, overal_cost(cost_loop_P,cost_loop_M,ratio,self.size_adj),cost_loop_P_dict,cost_loop_M_dict))

                if type=="loop_tau":
                    cost_loop_P_dict = cost_functions.cost_loop_tau(self.netP, sup, self.start_activities,self.end_activities)
                    cost_loop_P = sum(x['missing']+x['deviating'] for x in cost_loop_P_dict.values())
                    cost_loop_M_dict = cost_functions.cost_loop_tau(self.netM, sup, self.start_activitiesM,self.end_activitiesM)
                    cost_loop_M = sum(x['missing']+x['deviating'] for x in cost_loop_M_dict.values())
                    cut.append(((set(self.start_activities.keys()), set(self.end_activities.keys())), 'loop_tau', cost_loop_P,cost_loop_M, overal_cost(cost_loop_P,cost_loop_M,ratio,self.size_adj), cost_loop_P_dict,cost_loop_M_dict))


            if not cut:
                print("no good cut exists")
            sorted_cuts = sorted(cut, key=lambda x: (x[4], x[2],['exc_tau','exc','seq','par','loop','loop_tau'].index(x[1]), abs(len(x[0][0])-len(x[0][1]))))
            cut = sorted_cuts[0]

        # print(cut[:-2])

        map_cut_op = {'par': 'parallel', 'seq': 'sequential', 'exc': 'xor', 'exc_tau':'xor', 'exc2': 'xor',
                      'loop': 'loopCut', 'loop1': 'loopCut', 'loop_tau': 'loopCut'}

        if cut[1] in {'par','seq','exc','exc_tau','exc2','loop','loop_tau', 'single_activity'}:
            # updates = {'cut_type':cut[1],'set1':list(cut[0][0]), 'set2':list(cut[0][1]), 'cost+:':cut[2], 'cost-':ratio*self.size_adj*cut[3], 'contribusions+':{x:cut[5][x] for x in cut[5] if cut[5][x]>0} , 'contributions-':{x:(ratio*self.size_adj*cut[6][x]) for x in cut[6] if cut[6][x]>0},'overal_cost':cut[4]}
            updates = {'cut_type': cut[1], 'set1': list(cut[0][0]), 'set2': list(cut[0][1]), 'cost+:': cut[2],
                       'cost-': ratio * self.size_adj * cut[3],
                       'contribusions+': {x: {'deviating': cut[5][x]['deviating'], 'missing': cut[5][x]['missing']} for
                                          x in cut[5] if (cut[5][x]['missing'] > 0 or cut[5][x]['deviating'])},
                       'contributions-': {x: {'deviating': ratio * self.size_adj * cut[6][x]['deviating'],
                                              'missing': ratio * self.size_adj * cut[6][x]['missing']} for x in cut[6]
                                          if (cut[6][x]['missing'] > 0 or cut[6][x]['deviating'])},
                       'overal_cost': cut[4]}
            file_path = f'E:\\PADS\Projects\\IMbi_paper_revision\\experiments\\models\\data_s{int(10*sup)}_r{int(10*ratio)}.json'
            # read_append_write_json(file_path, updates)


        if cut[1] in map_cut_op.keys():
            self.detected_cut = map_cut_op[cut[1]]
            LAP, LBP = split.split(cut[1], [cut[0][0], cut[0][1]], self.log)
            LAM, LBM = split.split(cut[1], [cut[0][0], cut[0][1]], self.logM)
            new_logs = [[LAP, LAM], [LBP, LBM]]
            for l in new_logs:
                self.children.append(
                    SubtreePlain(l[0], l[1],
                                 self.rec_depth + 1,
                                 noise_threshold=self.noise_threshold,
                                 parameters=parameters, sup=sup, ratio=ratio, size_par=size_par, rules=rules))
        elif cut[1]!='single_activity' and cut[1]!='empty_log':
            print('It should not happen, if you see this error there could be a bug in the code!')




import networkx as nx
import json

def n_edges(net, S, T):
    edges_reweight = list(nx.edge_boundary(net, S, T, data='weight', default=1))
    return sum(weight for u, v, weight in edges_reweight if (u in S and v in T))


def get_edge_weight(graph, node1, node2):
    edge_data = graph.get_edge_data(node1, node2)
    if edge_data is None:
        return 0
    else:
        return edge_data['weight']

def add_SE(nodes, st,en):
    if nodes & set(st.keys()):
        nodes.add('start')
    if nodes & set(en.keys()):
        nodes.add('end')
    return nodes


def aggregate_dictionaries(deviating, missing):
    aggregated_dict = {}

    for d in deviating:
        for key, value in d.items():
            if key in aggregated_dict:
                aggregated_dict[key]['deviating'] += value
            else:
                aggregated_dict[key] = {'deviating':value,'missing':0}

    for d in missing:
        for key, value in d.items():
            if key in aggregated_dict:
                aggregated_dict[key]['missing'] += value
            else:
                aggregated_dict[key] = {'deviating':0,'missing':value}
    return aggregated_dict


def generate_nx_graph_from_log(log,nt):
    G = nx.DiGraph()
    for trace in log:
        tr_art = ('start',) + trace + ('end',)
        for i in range(len(tr_art) - 1):
            if G.has_edge(tr_art[i], tr_art[i + 1]):
                G[tr_art[i]][tr_art[i + 1]]['weight'] += log[trace]
            else:
                G.add_edge(tr_art[i], tr_art[i + 1], weight=log[trace])

    st = {}
    en = {}
    # Filter outgoing edges from start_node based on weight
    outgoing_edges = list(G.edges('start', data=True))
    for u, v, data in outgoing_edges:
        if data['weight'] < nt:
            G.remove_edge(u, v)
            if not nx.has_path(G, 'start', v):
               G.add_edge(u, v, weight=data['weight'])
               st[v] = data['weight']
        else:
            st[v] = data['weight']

    # Filter incoming edges to end_node based on frequency
    incoming_edges = list(G.in_edges('end', data=True))
    for u, v, data in incoming_edges:
        if data['weight'] < nt:
            G.remove_edge(u, v)
            if not nx.has_path(G, u, 'end'):
               G.add_edge(u, v, weight=data['weight'])
               en[u] = data['weight']
        else:
            en[u] = data['weight']
    return G

def generate_nx_indirect_graph_from_log(log):
    G = nx.DiGraph()
    for trace in log:
        for i in range(0,len(trace)):
            visited = set()
            for j in range(i+1,len(trace)):
                if trace[j] not in visited:
                    visited.add(trace[j])
                    if G.has_edge(trace[i], trace[j]):
                        G[trace[i]][trace[j]]['weight'] += log[trace]
                    else:
                        G.add_edge(trace[i], trace[j], weight=log[trace])
    return G

def read_append_write_json(file_path, new_item):
    # Read the current content of the JSON file
    with open(file_path, 'r') as file:
        data = json.load(file)

    # Append the new item to the list
    data.append(new_item)

    # Write the updated list back to the JSON file
    with open(file_path, 'w') as file:
        json.dump(data, file, indent=4)





import redis
import os

# Initialize Redis client
redis_client = redis.StrictRedis(
    host=os.getenv('REDIS_HOST', 'localhost'),
    port=6379,
    decode_responses=True
)

# Test connection (optional)
try:
    redis_client.ping()
    print("Connected to Redis!")
except redis.ConnectionError as e:
    print(f"Failed to connect to Redis: {e}")



from automata.fa.nfa import NFA
from automata.fa.dfa import DFA
from frozendict import frozendict

def extract_templates():
    return{
        "AtLeast1": (lambda a, S: f"{g(S-{a})}*({a}{g(S-{a})}*){{1,}}{g(S-{a})}*",1),
        "AtLeast2": (lambda a, S: f"{g(S-{a})}*({a}{g(S-{a})}*){{2,}}{g(S-{a})}*",1),
        "AtLeast3": (lambda a, S: f"{g(S-{a})}*({a}{g(S-{a})}*){{3,}}{g(S-{a})}*",1),
        "AtMost1": (lambda a, S: f"{g(S-{a})}*({a}{g(S-{a})}*){{0,1}}{g(S-{a})}*",1),
        "AtMost2": (lambda a, S: f"{g(S-{a})}*({a}{g(S-{a})}*){{0,2}}{g(S-{a})}*",1),
        "AtMost3": (lambda a, S: f"{g(S-{a})}*({a}{g(S-{a})}*){{0,3}}{g(S-{a})}*",1),
        "Absence": (lambda a, S: f"{g(S-{a})}*",1),
        "Init": (lambda a, S: f"{a}.*",1),
        "End": (lambda a, S: f".*{a}",1),
        "CoExistence": (lambda a, b, S: f"{g(S-{a,b})}*(({a}.*{b}.*)|({b}.*{a}.*))*{g(S-{a,b})}*",2),
        "RespondedExistence": (lambda a, b, S: f"{g(S-{a})}*(({a}.*{b}.*)|({b}.*{a}.*))*{g(S-{a})}*",2),
        "Response": (lambda a, b, S: f"{g(S-{a})}*({a}.*{b})*{g(S-{a})}*",2),
        "AlternateResponse": (lambda a, b, S: f"{g(S-{a})}*({a}{g(S-{a})}*{b}{g(S-{a})}*)*{g(S-{a})}*",2),
        "ChainResponse": (lambda a, b, S: f"{g(S-{a})}*({a}{b}{g(S-{a})}*)*{g(S-{a})}*",2),
        "Precedence": (lambda a, b, S: f"{g(S-{b})}*({a}.*{b})*{g(S-{b})}*",2),
        "AlternatePrecedence": (lambda a, b, S: f"{g(S-{b})}*({a}{g(S-{b})}*{b}{g(S-{b})}*)*{g(S-{b})}*",2),
        "ChainPrecedence": (lambda a, b, S: f"{g(S-{b})}*({a}{b}{g(S-{b})}*)*{g(S-{b})}*",2),
        "Succession": (lambda a, b, S: f"{g(S-{a,b})}*({a}.*{b})*{g(S-{a,b})}*",2),
        "AlternateSuccession": (lambda a, b, S: f"{g(S-{a,b})}*({a}{g(S-{a,b})}*{b}{g(S-{a,b})}*)*{g(S-{a,b})}*",2),
        "ChainSuccession": (lambda a, b, S: f"{g(S-{a,b})}*({a}{b}{g(S-{a,b})}*)*{g(S-{a,b})}*",2),
        "NotRespondedExistence": (lambda a, b, S: f"{g(S-{a,b})}*(({a}{g(S-{b})}*)|({b}{g(S-{a})}*))?",2),
        "NotCoExistence": (lambda a, b, S: f"{g(S-{a,b})}*(({a}{g(S-{b})}*)|({b}{g(S-{a})}*))?",2),
        "NotResponse": (lambda a, b, S: f"{g(S-{a})}*({a}{g(S-{b})}*)*{g(S-{a,b})}*",2),
        "NotChainResponse": (lambda a, b, S: f"{g(S-{a})}*({a}{a}*{g(S-{a,b})}{g(S-{a})}*)*({g(S-{a})}*|{a}*)",2),
        "NotPrecedence": (lambda a, b, S: f"{g(S-{a})}*({a}{g(S-{b})}*)*{g(S-{a,b})}*",2),
        "NotChainPrecedence": (lambda a, b, S: f"{g(S-{a})}*({a}{a}*{g(S-{a,b})}{g(S-{a})}*)*({g(S-{a})}*|{a}*)",2),
        "NotSuccession": (lambda a, b, S: f"{g(S-{a})}*({a}{g(S-{b})}*)*{g(S-{a,b})}*",2),
        # "NotChainSuccession": (lambda a, b, S: f"{g(S-{b})}*({b}{b}*{g(S-{a,b})}{g(S-{b})}*)*({g(S-{b})}*|{b}*)",2),
        "NotChainSuccession": (lambda a, b, S: f"{g(S-{a})}*({a}{a}*{g(S-{a,b})}{g(S-{a})}*)*({g(S-{a})}*|{a}*)",2),
    }
templates = extract_templates()

def g(input_set):
    """
    Converts a set of strings into a single string formatted as 'a|b|c'.

    Parameters:
    - input_set (set): A set of strings.

    Returns:
    - str: A single string in the format 'a|b|c'.
    """
    if not input_set:
        return ""  # Return an empty string for an empty set
    return f'({"|".join(sorted(input_set))})'

def gen_reg_dfa(template,parameters,S_mapped):
    S = set(S_mapped.values())
    if template in templates:

        if "Absence" in template or "Init" in template or "End" in template or "AtMost" in template or "AtLeast" in template:
            a = S_mapped[parameters[0]]
            reg = templates[template][0](a, S)
        else:
            a, b = S_mapped[parameters[0]], S_mapped[parameters[1]]
            reg = templates[template][0](a, b, S)
        # print(f"Generated regex for {template}: {reg}")
        nfa = NFA.from_regex(reg, input_symbols=S)
        dfa = rename_dfa_transitions(DFA.from_nfa(nfa), S_mapped)
        # Make the DFA complete
        complete_dfa = make_dfa_complete(dfa)
        return complete_dfa
    else:
        return ValueError("Invalid template")

def rename_dfa_transitions(dfa, mapping):
    """
    Renames the transition symbols in a DFA using a provided mapping dictionary.

    Parameters:
    - dfa (DFA): The input DFA whose transitions need renaming.
    - mapping (dict): A dictionary mapping current symbols to new symbols.

    Returns:
    - DFA: A new DFA with renamed transitions.
    """
    from frozendict import frozendict
    # from automata.fa.dfa import DFA
    mapping = {value: key for key, value in mapping.items()}
    # Rename the transitions
    renamed_transitions = {}
    for state, transition_map in dfa.transitions.items():
        renamed_transitions[state] = {
            mapping.get(symbol, symbol): dest_state  # Rename symbols based on mapping
            for symbol, dest_state in transition_map.items()
        }

    # Create a new DFA with renamed transitions
    return DFA(
        states=frozenset(dfa.states),
        input_symbols=frozenset(mapping.get(symbol, symbol) for symbol in dfa.input_symbols),
        transitions=frozendict({
            state: frozendict(trans) for state, trans in renamed_transitions.items()
        }),
        initial_state=dfa.initial_state,
        final_states=frozenset(dfa.final_states), allow_partial=True
    )

def make_dfa_complete(dfa):
    """
    Ensure the given DFA is complete by adding a dead state for missing transitions.
    """
    # Convert frozenset of states to a mutable set
    states = set(dfa.states)  # Convert frozenset to set for modification
    input_symbols = dfa.input_symbols  # Input symbols remain the same
    transitions = {state: dict(trans) for state, trans in dfa.transitions.items()}  # Convert frozendict to dict
    initial_state = dfa.initial_state  # Initial state remains the same
    final_states = dfa.final_states  # Final states remain the same

    # Add a dead state
    dead_state = "dead"
    states.add(dead_state)

    # Add missing transitions for each state
    for state in states:
        if state not in transitions:
            transitions[state] = {}
        for symbol in input_symbols:
            if symbol not in transitions[state]:
                transitions[state][symbol] = dead_state

    # Add self-loops on the dead state
    transitions[dead_state] = {symbol: dead_state for symbol in input_symbols}

    # Return the completed DFA with the updated states and transitions
    return DFA(
        states=frozenset(states),  # Convert back to frozenset for DFA consistency
        input_symbols=input_symbols,
        transitions=frozendict({state: frozendict(trans) for state, trans in transitions.items()}),  # Re-freeze transitions
        initial_state=initial_state,
        final_states=final_states,
    )


def assign_alphabet(S):
    """
    Assigns a unique alphabet character to each string in the list S.

    Parameters:
    - S (list): List of strings.

    Returns:
    - dict: A dictionary mapping each string in S to a unique alphabet character.
    """
    # List of single-character alphabet symbols
    alphabet = [chr(i) for i in range(ord('a'), ord('z') + 1)]  # 'a' to 'z'

    if len(S) > len(alphabet):
        raise ValueError("Not enough unique alphabet characters to assign to all strings in S.")

    # Map each string in S to a unique alphabet character
    mapping = {string: alphabet[i] for i, string in enumerate(S)}
    return mapping

def symbol_extract(tuple_str):
    tuple_str = tuple_str.strip("()")
    elements = tuple_str.split(", ")
    second_element = elements[1].strip("'")
    return second_element

def reachability2NFA(rg,activities):
    start_state = [s.name for s in rg.states if not s.incoming][0]
    final_state = set([s.name for s in rg.states if not s.outgoing])
    states = set([s.name for s in rg.states])
    input_symbols = activities
    # input_symbols = set()
    # for t in rg.transitions:
    #     if t.name=='None':
    #         input_symbols.add("")
    #     else:
    #         input_symbols.add(symbol_extract(t.name))
    transitions = {}
    for state in rg.states:
        tr_to_state = {}
        for tr in state.outgoing:
            if symbol_extract(tr.name) == 'None':
                if "" in tr_to_state:
                    tr_to_state[""].add(tr.to_state.name)
                else:
                    tr_to_state[""] = {tr.to_state.name}
            else:
                tr_to_state[symbol_extract(tr.name)] = {tr.to_state.name}
        transitions[state.name] = tr_to_state

    nfa = NFA(
            states=states,
            input_symbols=input_symbols,
            transitions=transitions,
            initial_state=start_state,
            final_states=final_state,
        )
    return nfa

def check_all(S1,S2,SS1, SS2, dfa_decl, cut_type, S_mapped):
    flag = False
    if cut_type == 'seq':
        for p in SS1:
            for q in SS2:
                reg_to_check1 = f"{g(S1 - {p})}*{p}{g(S1 - {p})}*{g(S2 - {q})}*{q}{g(S2 - {q})}*"
                nfa1 = NFA.from_regex(reg_to_check1, input_symbols=set(S_mapped.values()))
                dfa_to_check1 = rename_dfa_transitions(DFA.from_nfa(nfa1),S_mapped)
                if dfa_to_check1.issubset(dfa_decl.complement()):
                    flag = True
                    # block = True
                    break

    elif cut_type == 'exc':
        for p in SS1:
            reg_to_check2 = f"{g(S1 - {p})}*{p}{g(S1 - {p})}*"
            nfa2 = NFA.from_regex(reg_to_check2, input_symbols=set(S_mapped.values()))
            dfa_to_check2 = rename_dfa_transitions(DFA.from_nfa(nfa2),S_mapped)
            if dfa_to_check2.issubset(dfa_decl.complement()):
                flag = True
                # block = True
                break

        for q in SS2:
            reg_to_check3 = f"{g(S2 - {q})}*{q}{g(S2 - {q})}*"
            nfa3 = NFA.from_regex(reg_to_check3, input_symbols=set(S_mapped.values()))
            dfa_to_check3 = rename_dfa_transitions(DFA.from_nfa(nfa3),S_mapped)
            if dfa_to_check3.issubset(dfa_decl.complement()):
                # print(f"q is {q}")
                flag = True
                break



    elif cut_type == 'par':
        for p in SS1:
            for q in SS2:
                reg_to_check4 = f"{g(S1.union(S2) - {p, q})}*{p}{g(S1.union(S2) - {p, q})}*{q}{g(S1.union(S2) - {p, q})}*"
                reg_to_check5 = f"{g(S1.union(S2) - {p, q})}*{q}{g(S1.union(S2) - {p, q})}*{p}{g(S1.union(S2) - {p, q})}*"
                nfa4 = NFA.from_regex(reg_to_check4, input_symbols=set(S_mapped.values()))
                dfa_to_check4 = rename_dfa_transitions(DFA.from_nfa(nfa4),S_mapped)
                nfa5 = NFA.from_regex(reg_to_check5, input_symbols=set(S_mapped.values()))
                dfa_to_check5 = rename_dfa_transitions(DFA.from_nfa(nfa5),S_mapped)
                if dfa_to_check4.issubset(dfa_decl.complement()):
                    flag = True
                    # block = True
                    break
                if dfa_to_check5.issubset(dfa_decl.complement()):
                    flag = True
                    break

    elif cut_type == 'loop':
        for p in SS1:
            reg_to_check6 = f"{g(S1)}*{p}{g(S1)}*"
            nfa6 = NFA.from_regex(reg_to_check6, input_symbols=set(S_mapped.values()))
            dfa_to_check6 = rename_dfa_transitions(DFA.from_nfa(nfa6), S_mapped)
            if dfa_to_check6.issubset(dfa_decl.complement()):
                flag = True
                break
            for pp in SS1-{p}:
                reg_to_check9 = f"({g(S1-{p})}*{p}{g(S1-{p})}*)({g(S2)}*({g(S1-{pp})}*{pp}{g(S1-{pp})}))+"
                nfa9 = NFA.from_regex(reg_to_check9, input_symbols=set(S_mapped.values()))
                dfa_to_check9 = rename_dfa_transitions(DFA.from_nfa(nfa9), S_mapped)
                if dfa_to_check9.issubset(dfa_decl.complement()):
                    flag = True
                    break
            for q in SS2:
                reg_to_check7 = f"({g(S1-{p})}*{p}{g(S1-{p})}*)(({g(S2-{q})}*{q}{g(S2-{q})}*){g(S1)}*)+"
                nfa7 = NFA.from_regex(reg_to_check7, input_symbols=set(S_mapped.values()))
                dfa_to_check7 = rename_dfa_transitions(DFA.from_nfa(nfa7),S_mapped)
                if dfa_to_check7.issubset(dfa_decl.complement()):
                    flag = True
                    break
                reg_to_check8 = f"{g(S1)}*(({g(S2 - {q})}*{q}{g(S2 - {q})}*)({g(S1-{p})}*{p}{g(S1-{p})}*))+"
                nfa8 = NFA.from_regex(reg_to_check8, input_symbols=set(S_mapped.values()))
                dfa_to_check8 = rename_dfa_transitions(DFA.from_nfa(nfa8), S_mapped)
                if dfa_to_check8.issubset(dfa_decl.complement()):
                    flag = True
                    break
                for qq in SS1 - {q}:
                    reg_to_check10 = f"{g(S1)}*({g(S2-{q})}*{q}{g(S2-{q})}*){g(S1)}*(({g(S2-{qq})}*{qq}{g(S2-{qq})}*){g(S1)}*)+"
                    nfa10 = NFA.from_regex(reg_to_check10, input_symbols=set(S_mapped.values()))
                    dfa_to_check10 = rename_dfa_transitions(DFA.from_nfa(nfa10), S_mapped)
                    if dfa_to_check10.issubset(dfa_decl.complement()):
                        flag = True
                        break



    elif cut_type == 'loop_tau':
        for p in SS1:
            # p = S_mapped[pp]
            reg_to_check = f"{g(S1)}*{p}{g(S1)}"
            reg_to_check2 = f"({g(S1)}*{p}{g(S1)}){{2,}}"

            # reg_to_check2 = f"{g(S1)}*(({g(S2)}*{q}{g(S2)}*)({g(S1)}*{p}{g(S1)}*)){{2,}}"
            nfa = NFA.from_regex(reg_to_check, input_symbols=set(S_mapped.values()))
            dfa_to_check = rename_dfa_transitions(DFA.from_nfa(nfa),S_mapped)
            if dfa_to_check.issubset(dfa_decl.complement()):
                flag = True
                # block = True
                break
            nfa2 = NFA.from_regex(reg_to_check2, input_symbols=set(S_mapped.values()))
            dfa_to_check2 = rename_dfa_transitions(DFA.from_nfa(nfa2), S_mapped)
            if dfa_to_check2.issubset(dfa_decl.complement()):
                flag = True
                # block = True
                break
    elif cut_type == 'exc_tau':
        for p in SS1:
            reg_to_check = f"{g(S1 - {p})}*{p}{g(S1 - {p})}*"
            nfa = NFA.from_regex(reg_to_check, input_symbols=set(S_mapped.values()))
            dfa_to_check = rename_dfa_transitions(DFA.from_nfa(nfa), S_mapped)
            if dfa_to_check.issubset(dfa_decl.complement()):
                # print(f"p is {p}")
                # print(reg_to_check)
                flag = True
                # block = True
                break

        reg_to_check2 = ""
        nfa2 = NFA.from_regex(reg_to_check2, input_symbols=set(S_mapped.values()))
        dfa_to_check2 = rename_dfa_transitions(DFA.from_nfa(nfa2), S_mapped)
        if dfa_to_check2.issubset(dfa_decl.complement()):
            # print(f"p is {p}")
            # print(reg_to_check)
            flag = True
    return flag


def check_all_single(activity, dfa_decl, cut_type, S_mapped):
    flag = False
    if cut_type == 'single_single':
        reg_to_check = f"{activity}"
        nfa = NFA.from_regex(reg_to_check, input_symbols=set(S_mapped.values()))
        dfa_to_check = rename_dfa_transitions(DFA.from_nfa(nfa), S_mapped)
        if dfa_to_check.issubset(dfa_decl.complement()):
            flag = True
    elif cut_type == 'xor_single':
        reg_to_check = f"{activity}"
        nfa = NFA.from_regex(reg_to_check, input_symbols=set(S_mapped.values()))
        dfa_to_check = rename_dfa_transitions(DFA.from_nfa(nfa), S_mapped)
        if dfa_to_check.issubset(dfa_decl.complement()):
            flag = True
        reg_to_check2 = ""
        nfa2 = NFA.from_regex(reg_to_check2, input_symbols=set(S_mapped.values()))
        dfa_to_check2 = rename_dfa_transitions(DFA.from_nfa(nfa2), S_mapped)
        if dfa_to_check2.issubset(dfa_decl.complement()):
            # print(f"p is {p}")
            # print(reg_to_check)
            flag = True
    elif cut_type == 'loop_single':
        reg_to_check = f"{activity}"
        nfa = NFA.from_regex(reg_to_check, input_symbols=set(S_mapped.values()))
        dfa_to_check = rename_dfa_transitions(DFA.from_nfa(nfa), S_mapped)
        if dfa_to_check.issubset(dfa_decl.complement()):
            flag = True
        reg_to_check2 = f"{activity}{activity}"
        nfa2 = NFA.from_regex(reg_to_check2, input_symbols=set(S_mapped.values()))
        dfa_to_check2 = rename_dfa_transitions(DFA.from_nfa(nfa2), S_mapped)
        if dfa_to_check2.issubset(dfa_decl.complement()):
            flag = True
        reg_to_check3 = f"{activity}{activity}{activity}"
        nfa3 = NFA.from_regex(reg_to_check3, input_symbols=set(S_mapped.values()))
        dfa_to_check3 = rename_dfa_transitions(DFA.from_nfa(nfa3), S_mapped)
        if dfa_to_check3.issubset(dfa_decl.complement()):
            flag = True


    return flag





from pm4py.algo.conformance.alignments.petri_net import algorithm as alignments
from pm4py.algo.evaluation.replay_fitness import algorithm as replay_fitness
from pm4py.algo.conformance.alignments.dfg import algorithm as dfg_alignment
from pm4py.algo.evaluation.precision import algorithm as precision_evaluator

def apply(LPlus, LMinus, dfg_net, sa, ea):
    measures = {}
    print('hi')
    alp = dfg_alignment.apply(LPlus, dfg_net, sa, ea)
    fp_inf = replay_fitness.evaluate(alp, variant=replay_fitness.Variants.ALIGNMENT_BASED)
    fp = fp_inf['averageFitness']
    roc_data = [('p',x['fitness']) for x in alp]


    # alp = alignments.apply_log(LPlus, net, i_m, i_f)
    # fp_inf = replay_fitness.evaluate(alp, variant=replay_fitness.Variants.ALIGNMENT_BASED)
    # fp = fp_inf['averageFitness']


    ################################################################################
    # prec_Plus = precision_evaluator.apply(LPlus, net, i_m, i_f,
    #                                       variant=precision_evaluator.Variants.ALIGN_ETCONFORMANCE)
    ################################################################################

    alm = dfg_alignment.apply(LMinus, dfg_net, sa, ea)
    fm_inf = replay_fitness.evaluate(alm, variant=replay_fitness.Variants.ALIGNMENT_BASED)
    fm = fm_inf['averageFitness']
    roc_data += [('n',x['fitness']) for x in alm]

    # alm = alignments.apply_log(LMinus, net, i_m, i_f)
    # fm_inf = replay_fitness.evaluate(alm, variant=replay_fitness.Variants.ALIGNMENT_BASED)
    # fm = fm_inf['averageFitness']


    measures['acc'] = round(fp - fm,2)
    measures['F1'] = round(2 * ((fp*(1-fm))/(fp+(1-fm))),2)
    # measures['precision'] = prec_Plus
    measures['fitP'] = round(fp,2)
    measures['fitM'] = round(fm,2)
    measures['roc_data'] = roc_data


    return measures


def apply_petri(LPlus, LMinus, net, i_m, i_f):
    measures = {}


    alp = alignments.apply_log(LPlus, net, i_m, i_f)
    fp_inf = replay_fitness.evaluate(alp, variant=replay_fitness.Variants.ALIGNMENT_BASED)
    fp = fp_inf['averageFitness']
    fp_pef = fp_inf['percentage_of_fitting_traces']/100
    # roc_data = [('p', x['fitness']) for x in alp]


    ################################################################################
    prec_Plus = precision_evaluator.apply(LPlus, net, i_m, i_f,
                                          variant=precision_evaluator.Variants.ALIGN_ETCONFORMANCE)
    ################################################################################


    alm = alignments.apply_log(LMinus, net, i_m, i_f)
    fm_inf = replay_fitness.evaluate(alm, variant=replay_fitness.Variants.ALIGNMENT_BASED)
    fm = fm_inf['averageFitness']
    fm_pef = fm_inf['percentage_of_fitting_traces']/100
    # roc_data += [('n', x['fitness']) for x in alm]


    measures['acc'] = round(fp - fm,2)
    measures['F1'] = round(2 * ((fp*(1-fm))/(fp+(1-fm))),2)
    measures['precision'] = round(prec_Plus,2)
    measures['fitP'] = round(fp,2)
    measures['fitM'] = round(fm,2)
    measures['acc_perf'] = round(fp_pef - fm_pef,2)
    measures['F1_perf'] = round(2 * ((fp_pef*(1-fm_pef))/(fp_pef+(1-fm_pef))),2)

    TP = fp_pef*len(LPlus)
    FP = fm_pef*len(LMinus)
    FN = (1-fp_pef)*len(LPlus)
    TN = (1-fm_pef)*len(LMinus)
    measures['acc_ML'] = (TP+TN)/(TP+TN+FP+FN)
    if (TP + FP) != 0:
        measures['prc_ML'] = TP / (TP + FP)
    else:
        measures['prc_ML'] = 'ignore'

    if (TP + FN) != 0:
        measures['rec_ML'] = TP / (TP + FN)
    else:
        measures['rec_ML'] = 'ignore'
    # measures['roc_data'] = roc_data


    return measures



